/*
 * For erasing a flash partition.
 *
 * (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <linux/version.h>
#ifdef __elinux__
#include <linux/flash.h>
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 6, 8)
#include <mtd/mtd-user.h>
#else
#include <linux/mtd/mtd.h>
#endif
#endif
#include <sys/types.h>
#include <sys/ioctl.h>
#include <syslog.h>



/* Show usage.                 */
void usage(void) 
{
        fprintf(stderr, "Usage: eraseflash [OPTION] device\n"
#ifdef __elinux__
                "Erase a whole flash device.\n"
                "\n"
#else
                "Erase a whole, or part of, a flash device.\n"
                "\n"
                "  -s, --sector <sector>  Start sector for erase\n"
                "  -c, --count  <count>   Number of sectors to erase\n"
#endif
                "  -v, --verbose          Emit verbose messages\n"
                "  -h, --help             This help\n"
                "\n"
#ifndef __elinux__
                "With -s, -c must also be specified.  Without either -s or -c the whole device\n"
                "is erased.\n"
#endif
                );
}

#ifndef __elinux__
/* Check sector number. Sector numbering is 0-based. */
int sector_valid(int sector)
{
        return (sector >= 0);
}
#endif

/* Check sector count. Sector count must be positive. */
#ifndef __elinux__
int count_valid(int count)
{
        return (count > 0);
}
#endif


/* Erase the flash device. */
#ifdef __elinux__
int erase_flash(int fd)
#else
int erase_flash(int fd, int sector, int count, int verbose)
#endif
{
#ifdef __elinux__
        if (ioctl(fd, FLASHIO_ERASEALL, NULL) < 0) {
                perror("ioctl FLASHIO_ERASEALL");
                return 1;
        }
#else
        mtd_info_t mtd_info;
        erase_info_t mtd_erase;
	int blockbad = 0;
	int s;

        if ((ioctl(fd, MEMGETINFO, &mtd_info)) < 0) {
                perror("ioctl mtd MEMGETINFO");
                return 1;
        }

	if (sector_valid(sector) && count_valid(count)) {
		/* Sanity check. */
		if (sector * mtd_info.erasesize + 
		    count * mtd_info.erasesize > mtd_info.size) {
		  syslog(LOG_WARNING, "Attempt to erase flash, but "
			 "sector and count exceeds flash size.");
		  fprintf(stderr, "Error: sector and count"
			  " exceeds flash size\n");
		  return 1;
		}
	} else {
		/* No sector or count; erase the whole device. */
		sector = 0;
		count = mtd_info.size / mtd_info.erasesize;
	}

	if (verbose)
		printf("Erasing flash of length %u at offset 0x%x\n",
			count, sector);
	for (s = 0; s < count; s++) {
		mtd_erase.start = (s + sector) * mtd_info.erasesize;
		mtd_erase.length = mtd_info.erasesize;
		loff_t offs = mtd_erase.start;
		if (mtd_info.type == MTD_NANDFLASH &&
		    (blockbad = ioctl(fd, MEMGETBADBLOCK, &offs)) < 0) {
			perror("ioctl mtd MEMGETBADBLOCK");
			return 1;
		}
		if (blockbad) {
			if (verbose)
				printf("Block %d at 0x%x bad (%d), skipping\n",
				       s + sector, mtd_erase.start, blockbad);
			continue;
		}
		if ((ioctl(fd, MEMERASE, &mtd_erase)) < 0) {
			if (mtd_info.type == MTD_NANDFLASH && errno == EIO) {
				if (verbose)
					printf("Block %d at 0x%x now bad, skipping\n",
					       s + sector, mtd_erase.start);
				if (ioctl(fd, MEMSETBADBLOCK, &offs) < 0) {
					perror("ioctl mtd MEMSETBADBLOCK");
					return 1;
				}
			} else {
				perror("ioctl mtd MEMERASE");
				return 1;
			}
		}
        }
#endif
        return 0;
}

/* The main() function:                */
int main(int argc, char **argv)
{
	int ret = 0;
	int verbose = 0;
	char *device_name = 0;
        
	int fd;

        /* Getopt stuff. */
        int c;
        int option_index;

#ifdef __elinux__
        char options[] = "vh";
#else
        /* Initialize with invalid values. */
        int sector = -1;
        int count = 0;
        char options[] = "vs:c:h";
#endif

        static struct option long_options[] =
        {
                {"verbose", no_argument, 0, 'v'},
#ifndef __elinux__
                {"sector", required_argument, 0, 's'},
                {"count", required_argument, 0, 'c'},
#endif
                {"help", no_argument, 0, 'h'},
                {0, 0, 0, 0}
        };

        if(argc < 2)
        {
                usage();
                return 1;
        }

        openlog("eraseflash", LOG_PID, LOG_USER);
        while((c = getopt_long (argc, argv, options, 
                                 long_options, &option_index)) != -1)
        {
                switch(c)
                {
                case 'v':
                        verbose = 1;
                        break;
#ifndef __elinux__
                case 's':
                        sector = atoi(optarg);
                        if (!sector_valid(sector)) {
                                fprintf(stderr, "Error: invalid sector\n");
                                syslog(LOG_WARNING, "Attempt to erase flash with invalid sector specified.");
                                return 1;
                        }
                        break;

                case 'c':
                        count = atoi(optarg);
                        if (!count_valid(count)) {
                                fprintf(stderr, "Error: invalid count\n");
                                syslog(LOG_WARNING, "Attempt to erase flash with invalid count specified.");
                                return 1;
                        }
                        break;
#endif
                case 'h':
                        usage();
                        return 0;

                default:
                        usage();
                        return 1;
                }
        }

#ifndef __elinux__
        /* Quick sanity check that not just one of sector and count is
           specified. */
        if ((sector_valid(sector) && !count_valid(count)) ||
            (!sector_valid(sector) && count_valid(count))) {
                fprintf(stderr, "Error: both sector and count must be"
                        " specified\n");
                syslog(LOG_WARNING, "Attempt to erase flash with missing count or sector specification.");
                return 1;
        }
#endif

        /* optind is the index in argv of the first argv element that is
           not an option, in this case the device name. */
        device_name = argv[optind];
        if (!device_name) {
                fprintf(stderr, "Error: no device specified\n");
                syslog(LOG_WARNING, "Attempt to erase flash with missing device specification.");
                return 1;
        }
        
        if ((fd = open(device_name, O_RDWR)) < 0) {
		perror("eraseflash: open");
                syslog(LOG_ERR, "Attempt to erase flash but not possible to open device.");
		return 1;
	}

#ifdef __elinux__
	if (verbose) {
		printf("Erasing flash %s...", device_name);
		fflush(stdout);
	}
        ret = erase_flash(fd);
#else
        ret = erase_flash(fd, sector, count, verbose);
#endif

	close(fd);

	if (verbose) {
		printf("Flash erasure %s\n", (ret)? "failed!" : "OK");
	}
        syslog(LOG_INFO, "Flash erasure %s", (ret)? "failed!" : "OK.");

	return ret;
}
