/*
 * Display info about flash device
 *
 * (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 6, 8)
#include <mtd/mtd-user.h>
#else
#include <linux/mtd/mtd.h>
#endif
#include <sys/types.h>
#include <sys/ioctl.h>
#include <syslog.h>



/* Show usage.                 */
void usage(void) 
{
        fprintf(stderr, "Usage: flashinfo [OPTION] device\n"
                "Show info about flash device.\n"
                "\n"
                "  -h, --help             This help\n");
}

/* return string with flash type info */
char *flash_type(int type)
{
	switch (type) {
		case MTD_ABSENT: return "ABSENT";
		case MTD_RAM: return "RAM";
		case MTD_ROM: return "ROM";
		case MTD_NORFLASH: return "NOR FLASH";
		case MTD_NANDFLASH: return "NAND FLASH";
		case MTD_DATAFLASH: return "DATA FLASH";
#if LINUX_VERSION_CODE < KERNEL_VERSION (2, 6, 18)
		case MTD_PEROM: return "PEROM";
		case MTD_OTHER: return "OTHER";
		case MTD_UNKNOWN: return "UNKNOWN";
#endif
		default: break;
	}
	return "unknown";
}
				 

/* Show info about flash device. */
int show_info(int fd, int verbose)
{
        mtd_info_t mtd_info;
	struct nand_oobinfo oobinfo;
	int bytecount;
	int i;

        if ((ioctl(fd, MEMGETINFO, &mtd_info)) < 0) {
                perror("ioctl mtd MEMGETINFO");
                return 1;
        } else {
		printf("Type %s (%d), flags 0x%x\n", flash_type(mtd_info.type), 
		       mtd_info.type, mtd_info.flags);
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 6, 18)
		/* most of the flags have been removed in 2.6.18, and 
		 * are replaced with the simple ones below.
		 */
		printf("%s ", mtd_info.flags & MTD_WRITEABLE ? "WRITEABLE" : "writeable");
		printf("%s ", mtd_info.flags & MTD_BIT_WRITEABLE ? "BIT_WRITEABLE" : "bit_writeable");
		printf("%s ", mtd_info.flags & MTD_NO_ERASE ? "NO_ERASE" : "no_erase");
#else
		/* mtd in older kernels have low-level flags */
		printf("%s ", mtd_info.flags & MTD_CLEAR_BITS ? "CLEAR_BITS" : "clear_bits");
		printf("%s ", mtd_info.flags & MTD_SET_BITS ? "SET_BITS" : "set_bits");
		printf("%s ", mtd_info.flags & MTD_ERASEABLE ? "ERASEABLE" : "eraseable");
		printf("%s ", mtd_info.flags & MTD_WRITEB_WRITEABLE ? "WRITEB_WRITEABLE" : "writeb_writable");
		printf("%s\n", mtd_info.flags & MTD_VOLATILE ? "VOLATILE" : "volatile");
		printf("%s ", mtd_info.flags & MTD_XIP ? "XIP" : "xip");
		printf("%s ", mtd_info.flags & MTD_OOB ? "OOB" : "oob");
		printf("%s ", mtd_info.flags & MTD_ECC ? "ECC" : "ecc");
		printf("%s ", mtd_info.flags & MTD_NO_VIRTBLOCKS ? "NO_VIRTBLOCKS" : "no_virtblocks");
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 6, 16)
		printf("%s", mtd_info.flags & MTD_PROGRAM_REGIONS ? "PROGRAM_REGIONS" : "program_regions");
#endif /* 2.6.16 */
#endif /* 2.6.18 */
		printf("\n");
		printf("Total size %d (0x%x), erasesize %d (0x%x)\n",
		       mtd_info.size, mtd_info.size,
		       mtd_info.erasesize, mtd_info.erasesize);
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 6, 18)
		printf("write size %d (0x%x), oob size %d (0x%x)\n",
		       mtd_info.writesize, mtd_info.writesize,
		       mtd_info.oobsize, mtd_info.oobsize);
#else
		printf("oob block size %d (0x%x), oob data per block %d (0x%x)\n",
		       mtd_info.oobblock, mtd_info.oobblock,
		       mtd_info.oobsize, mtd_info.oobsize);
#endif
		printf("ecc type %d, ecc size %d (0x%x)\n",
		       mtd_info.ecctype, mtd_info.eccsize, mtd_info.eccsize);

        }

	int region_count;

	if (ioctl(fd, MEMGETREGIONCOUNT, &region_count) < 0) {
		perror("ioctl mtd MEMGETREGIONCOUNT");
		return 1;
	}
	if (region_count) {
		printf("Regions: %d\n", region_count);

		int region;
		struct region_info_user region_info;
		for (region = 0; region < region_count; region++) {
			region_info.regionindex = region;
			if (ioctl(fd, MEMGETREGIONINFO, &region_info) < 0) {
				perror("ioctl mtd MEMGETREGIONINFO");
				return 1;
			}
			printf("Region %d: start %d (0x%x), "
			       "erasesize %d (0x%x), blocks %d\n", region, 
			       region_info.offset, region_info.offset,
			       region_info.erasesize, region_info.erasesize, 
			       region_info.numblocks);
		}
	}

        unsigned int block_size = mtd_info.erasesize;
	unsigned int total_size = mtd_info.size;
	loff_t offs;

	if (mtd_info.type == MTD_NANDFLASH) {
#ifdef MTD_OOB
		if (mtd_info.flags & MTD_OOB) {
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 6, 18)
			printf("Info from obsolete MEMGETOOBSEL: ");
#endif
			if ((ioctl(fd, MEMGETOOBSEL, &oobinfo)) < 0) {
				perror("ioctl mtd MEMGETOOBSEL");
				return 1;
			}
			printf("ecc placement %u, ecc bytes %u\n",
			       oobinfo.useecc, oobinfo.eccbytes);
			printf("ecc oob allocation: ");
			bytecount = oobinfo.eccbytes;
			if (bytecount > 32)
				bytecount = 32; /* max size, see mtd-abi.h */
			for (i = 0; i < bytecount; i++)
				printf("%d ", oobinfo.eccpos[i]);
			printf("\n");
#ifdef MTD_OOB
		}
#endif

#ifdef ECCGETLAYOUT
		/* note: doesn't work on stock 2.6.18 due to bug in mtdchar.c */

		struct nand_ecclayout ecclayout;

		printf("Info from ECCGETLAYOUT: ");
		if ((ioctl(fd, ECCGETLAYOUT, &ecclayout)) < 0 ) {
			perror("ioctl mtd ECCGETLAYOUT");
			return 1;
		}
		printf("ecc bytes %u, oobavail %u\n", 
		       ecclayout.eccbytes, ecclayout.oobavail);
		printf("ecc oob allocation: ");
		bytecount = oobinfo.eccbytes;
		if (bytecount > 64)
			bytecount = 64; /* max size, see mtd-abi.h */
		for (i = 0; i < bytecount; i++)
			printf("%d ", ecclayout.eccpos[i]);
		printf("\n");
		printf("free oob bytes (offset,length): ");
		bytecount = ecclayout.oobavail;
		if (bytecount > MTD_MAX_OOBFREE_ENTRIES)
			bytecount = MTD_MAX_OOBFREE_ENTRIES;
		for (i = 0; i < bytecount; i++)
			printf("%d,%d; ", ecclayout.oobfree[i].offset, 
					  ecclayout.oobfree[i].length);
		printf("\n");
#endif

#ifdef ECCGETSTATS
		struct mtd_ecc_stats ecc_stats;
		printf("Info from ECCGETSTATS: ");
		if ((ioctl(fd, ECCGETSTATS, &ecc_stats)) < 0 ) {
			perror("ioctl mtd ECCGETSTATS");
			return 1;
		}
		printf("corrected %u, failed %u, bad blocks %u, bbt blocks %u\n",
		       ecc_stats.corrected, ecc_stats.failed,
		       ecc_stats.badblocks, ecc_stats.bbtblocks);
#endif

#ifdef MEMGETBADBLOCK
		/* check and display bad blocks */
		printf("Bad blocks:\n");
		for (offs = 0; offs < total_size; offs += block_size) {
			int blockbad;
			if ((blockbad = ioctl(fd, MEMGETBADBLOCK, &offs)) < 0) {
				perror("ioctl mtd MEMGETBADBLOCK");
				return 1;
			}
			if (blockbad)
				printf("Block at 0x%llx (block %llu) bad (%d).\n", offs, offs / block_size, blockbad);
			else if (verbose)
				printf("Block at 0x%llx (block %llu) good.\n", offs, offs / block_size);
		}
#endif
	}

        return 0;
}

/* The main() function:                */
int main(int argc, char **argv)
{
	int ret = 0;
	int verbose = 0;
	char *device_name = 0;
        
	int fd;

        /* Getopt stuff. */
        int c;
        int option_index;

        char options[] = "hv";

        static struct option long_options[] =
        {
                {"help", no_argument, 0, 'h'},
                {"verbose", no_argument, 0, 'v'},
                {0, 0, 0, 0}
        };

        if(argc < 2)
        {
                usage();
                return 1;
        }

        while((c = getopt_long (argc, argv, options, 
                                 long_options, &option_index)) != -1)
        {
                switch(c)
                {
                case 'h':
                        usage();
                        return 0;
		case 'v':
			verbose = 1;
			break;
                default:
                        usage();
                        return 1;
                }
        }

        /* optind is the index in argv of the first argv element that is
           not an option, in this case the device name. */
        device_name = argv[optind];
        if (!device_name) {
                fprintf(stderr, "Error: no device specified\n");
                return 1;
        }

	printf("Device: %s\n", device_name);
        
        if ((fd = open(device_name, O_RDONLY)) < 0) {
		perror("flashinfo: open");
		return 1;
	}

        ret = show_info(fd, verbose);

	close(fd);

	return ret;
}
