/*
 * Lock/unlouck flash device.
 *
 * (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 6, 8)
#include <mtd/mtd-user.h>
#else
#include <linux/mtd/mtd.h>
#endif
#include <sys/types.h>
#include <sys/ioctl.h>
#include <syslog.h>



/* Show usage.                 */
void usage(void) 
{
        fprintf(stderr, "Usage: flashinfo [OPTION] device\n"
                "Show info about flash device.\n"
                "\n"
		"  -l  --lock             Lock device\n"
		"  -u, --unlock           Unlock device\n"
                "  -h, --help             This help\n");
}


int main(int argc, char **argv)
{
	int verbose = 0;
	char *device_name = 0;
	int lock_device = 0, unlock_device = 0;
        mtd_info_t mtd_info;
	erase_info_t mtd_lock_info;
        
	int fd;

        /* Getopt stuff. */
        int c;
        int option_index;

        char options[] = "ulhv";

        static struct option long_options[] =
        {
                {"unlock", no_argument, 0, 'u'},
                {"lock", no_argument, 0, 'l'},
                {"help", no_argument, 0, 'h'},
                {"verbose", no_argument, 0, 'v'},
                {0, 0, 0, 0}
        };

        if(argc < 2)
        {
                usage();
                return 1;
        }

        while((c = getopt_long (argc, argv, options, 
                                 long_options, &option_index)) != -1)
        {
                switch(c)
                {
		case 'l':
			lock_device = 1;
			if (unlock_device) {
				usage();
				return 1;
			}
			break;
		case 'u':
			unlock_device = 1;
			if (lock_device) {
				usage();
				return 1;
			}
			break;
                case 'h':
                        usage();
                        return 0;
		case 'v':
			verbose = 1;
			break;
                default:
                        usage();
                        return 1;
                }
        }

	if (!lock_device && !unlock_device) {
		usage();
		return 1;
	}

        /* optind is the index in argv of the first argv element that is
           not an option, in this case the device name. */
        device_name = argv[optind];
        if (!device_name) {
                fprintf(stderr, "Error: no device specified\n");
                return 1;
        }

	printf("Device: %s\n", device_name);
        
        if ((fd = open(device_name, O_RDONLY)) < 0) {
		perror("flashinfo: open");
		return 1;
	}

        if ((ioctl(fd, MEMGETINFO, &mtd_info)) < 0) {
                perror("ioctl mtd MEMGETINFO");
                return 1;
        }

	mtd_lock_info.start = 0;
	mtd_lock_info.length = mtd_info.size;

	if (ioctl(fd, lock_device ? MEMLOCK : MEMUNLOCK, &mtd_lock_info) < 0) {
		perror("ioctl mtd MEM[UN]LOCK");
		return 2;
	}

	printf("Device %slocked!\n", lock_device ? "" : "un");

	close(fd);

	return 0;
}
