/* vi: set sw=2 ts=8: */
/*
 * flashnand - write image to nand flash, taking partition table 
 * into consideration
 *
 * (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "nandlib.h"

#define ROTATING_PROGRESS_INDICATOR (1)

const char program_name[] = "flashnand";

char *nand_device;
char *image_filename;

int verbose = 0;
int quiet = 0;

void progress_indicator(unsigned int percent)
{
#if ROTATING_PROGRESS_INDICATOR
  static int rotator;

  if (!quiet) {
    printf("%c\b", "|/-\\"[rotator++ & 3]);
    fflush(stdout);
  }
#else
  printf("Done (%d%%)\n", percent);
#endif
}


void 
display_help(void)
{
  printf("Usage:\n"
         "flashnand [options] <device> <image>\n"
         "Write <image> to nand flash memory <device>\n"
         "Options:\n"
         "-h, --help               this message\n"
         "-v, --verbose            verbose output\n"
         "-p, --part <partition>   start with partition <partition>\n"
         "-q, --quiet              no output (except for errors)\n");
}

int
main(int argc, char **argv)
{
  int start_partition = 0;
  int c;
  struct stat filestats;
  int image_size;
  unsigned char *filebuf = NULL;
  int fd;

  for (;;) {
    int option_index = 0;
    static const char *short_options = "hvqp:";
    static const struct option long_options[] = {
      {"help", no_argument, 0, 'h'},
      {"verbose", no_argument, 0, 'v'},
      {"quiet", no_argument, 0, 'q'},
      {"part", required_argument, 0, 'p'},
      {0, 0, 0, 0},
    };

    c = getopt_long(argc, argv, short_options, long_options, &option_index);
    if (c == EOF) {
      break;
    }

    switch (c) {
    case 'h':
    case '?':
      display_help();
      exit(0);
    case 'v':
      verbose = 1;
      quiet = 0;
      break;
    case 'q':
      quiet = 1;
      verbose = 0;
      break;
    case 'p':
      start_partition = atol(optarg);
      break;
    default:
      exit(1);
    }
  }
  
  if ((argc - optind) != 2) {
    display_help ();
    exit(1);
  }

  nandlib_set_verbosity(stdout, verbose, quiet);
  
  nand_device = argv[optind++];
  image_filename = argv[optind];

  if (verbose) {
    printf("NAND flash device to write to: %s\n", nand_device);
    printf("Image file to read from: %s\n", image_filename);
  }

  if (stat(image_filename, &filestats) < 0) {
    perror(program_name);
    exit(2);
  }

  image_size = filestats.st_size;
  if (verbose)
    printf("Image size is %d bytes\n", image_size);

  filebuf = malloc(image_size);
  if (!filebuf) {
    perror(program_name);
    exit(2);
  }

  fd = open(image_filename, O_RDONLY);
  if (fd < 0) {
    perror(program_name);
    exit(2);
  }

  if (read(fd, filebuf, image_size) != image_size) {
    printf("Couldn't read file\n");
    close(fd);
    exit(3);
  }

  close(fd);

  if (nandlib_check_image(filebuf, image_size, start_partition) < 0) {
    printf("Bad image, aborting.\n");
    exit(3);
  }

  /* flash */

  /* open the flash device */
  fd = open(nand_device, O_RDWR);
  if (fd < 0) {
    perror("open flash");
    exit(2);
  }

  if (nandlib_check_flash(fd) < 0)
    exit(2);

  /* first erase whole flash */

  if (!quiet)
    printf("Erasing flash...");

  if (nandlib_erase_flash(fd, start_partition, progress_indicator) < 0) {
    printf("Error erasing flash.\n");
    close(fd);
    exit(2);
  }

  if (!quiet)
    printf("done.\n");

  /* then program it */

  if (nandlib_program_flash(fd, start_partition, progress_indicator) < 0) {
    printf("Error programming flash.\n");
    close(fd);
    exit(2);
  }

  /* and finally verify it */

  if (nandlib_verify_flash(fd, start_partition, progress_indicator) < 0) {
    printf("Error verifying flash.\n");
    close(fd);
    exit(2);
  }


  close(fd);

  /* clean up */
  
  if (filebuf) {
    free(filebuf);
    filebuf = NULL;
  }
  if (!quiet) {
    printf("Done.\n");
  }
  return 0;
}
