/* vi: set sw=2 ts=8: */
/*
 * nandlib - routines for NAND flash erase, write, partition checks
 *
 * (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <mtd/mtd-user.h>
#include <string.h>
#include <errno.h>

#include <fwinfo.h>

#include "ptable.h"

#include "nandlib.h"

#ifndef MEMGETBADBLOCK
#error This program requires a newer kernel with MEMGETBADBLOCK in mtd!
#endif

/* For testing */
#undef FORCE_VERIFY_ERROR

static int quiet, verbose;

struct partition {
  unsigned long start_addr;
  unsigned long size;
  const unsigned char *buf;
  unsigned short flags;
  unsigned short type;
  unsigned long csum;
};

static struct partition partitions[MAX_ENTRIES+1];
static int nr_partitions;

const unsigned char *buf_max; /* end of buffer */

/* info about the flash chips in use */
static struct mtd_info_user meminfo;

/* info from the image to be written */
static struct fwinfo fwinfo;

static int flash_checked = 0;
static int image_checked = 0;

static FILE *of = NULL;


static void
indicate_progress(progressor progress, unsigned int howfar, unsigned int maxval)
{
  static unsigned int old_percent = 101; /* 101 forces if clause first time */
  /* /0x10000 to avoid overflow with large memories (32 bits overflow at 4GB) */
  unsigned int percent = (howfar / 0x10000) * 100 / (maxval / 0x10000);

  if (percent != old_percent)
  {
    old_percent = percent;
    progress(percent);
  }
}


static unsigned long
checksum(const unsigned char *addr, unsigned int length)
{
  int i;
  unsigned long csum = 0;

  for (i = 0; i < length; i++)
    csum += *addr++;
  return csum;
}


static void
show_partition(int partition_no, const struct partition *partition)
{
  fprintf(of, "Partition %d: start at 0x%08lx, size 0x%08lx, type 0x%04x, flags 0x%04x\n", 
          partition_no,
          partition->start_addr,
          partition->size,
          partition->type,
          partition->flags);
  fprintf(of, "             buf %p, csum %ld\n", 
          partition->buf,
          partition->csum);
  fprintf(of, "             contents: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x...\n",
          partition->buf[0],
          partition->buf[1],
          partition->buf[2],
          partition->buf[3],
          partition->buf[4]);
}


void 
nandlib_set_verbosity(FILE *out_f, int new_verbose, int new_quiet)
{
  of = out_f;
  verbose = new_verbose;
  quiet = new_quiet;
}


static int
read_partition_table(const unsigned char *filebuf)
{
  struct partitiontable *ptable = 
    (struct partitiontable *) (filebuf + fwinfo.ptable_img_addr);
  const unsigned char *buf = filebuf;
  int entries;
  struct partition *partition;
  int partition_no;

  if (verbose)
    fprintf(of, "File buffer at %p, partition table at %p\n",
            filebuf, ptable);

  if (ptable->head.magic != PARTITION_TABLE_MAGIC) {
    fprintf(of, "Bad partition table magic: 0x%x\n", ptable->head.magic);
    return -1;
  }

  entries = (ptable->head.size - PARTITIONTABLE_END_MARKER_SIZE) / 
             sizeof(struct partitiontable_entry);

  if (entries * sizeof(struct partitiontable_entry) + 
      PARTITIONTABLE_END_MARKER_SIZE != ptable->head.size) {
    fprintf(of, "Bad partition table size: %d\n", ptable->head.size);
    return -1;
  }

  if (entries > MAX_ENTRIES) {
    fprintf(of, "Too many entries (%d) in partition table (max %d)\n", 
            entries, MAX_ENTRIES);
    return -1;
  }

  if (ptable->entries[entries].offset != PARTITIONTABLE_END_MARKER) {
    fprintf(of, "End marker not found (0x%lx).\n", 
            ptable->entries[entries].offset);
    return -1;
  }

  if (checksum((const unsigned char *) &ptable->entries[0], 
               entries * sizeof(struct partitiontable_entry) + 
               PARTITIONTABLE_END_MARKER_SIZE) != 
      ptable->head.checksum) {
    fprintf(of, "Bad partition table checksum.\n");
    return -1;
  }

  if (!quiet)
    fprintf(of, "%d entries in partition table.\n", entries);

  /* rescue partition - default */
  partitions[0].start_addr = 0;
  partitions[0].size = fwinfo.ptable_img_addr;
  partitions[0].buf = buf;
  partitions[0].flags = PARTITION_FLAGS_READONLY;
  partitions[0].type = PARTITION_TYPE_RESCUE;

  if (verbose)
    show_partition(0, &partitions[0]);

  buf += fwinfo.ptable_img_addr;

  partition = &partitions[1];
  partition_no = 0;

  /* now read the table */
  while (partition_no <= entries) {
    if (ptable->entries[partition_no].offset == PARTITIONTABLE_END_MARKER)
      break;
    partition->start_addr = 
      ptable->entries[partition_no].offset + fwinfo.ptable_flash_addr;
    partition->size =
      ptable->entries[partition_no].size;
    partition->buf = buf;
    partition->flags = ptable->entries[partition_no].flags;
    partition->type = ptable->entries[partition_no].type;
    partition->csum = ptable->entries[partition_no].checksum;

    if (verbose)
      show_partition(partition_no + 1, partition);

    buf += partition->size;
    partition++;
    partition_no++;
  }

  nr_partitions = partition_no + 1;
  buf_max = buf;

  if (!quiet)
    fprintf(of, "%d partitions found.\n", nr_partitions);

  return 0;
}


static int 
check_partitions(const unsigned char *filebuf)
{
  int partition_no;
  int retval = 0;
  struct partition *partition;
  const unsigned char *buf;
  unsigned int size;
  unsigned int ptable_offset;

  /* Note: start at 1; rescue partition cannot be checked */
  for (partition_no = 1; partition_no < nr_partitions; partition_no++)
  {
    partition = &partitions[partition_no];
    buf = partition->buf;
    size = partition->size;
    if (partition_no == 1) {
      ptable_offset = PARTITION_TABLE_OFFSET + 
        sizeof(struct partitiontable_head) + 
        sizeof(struct partitiontable_entry) * 
        (nr_partitions - 1) + 
        PARTITIONTABLE_END_MARKER_SIZE +
        PARTITIONTABLE_END_PAD;
      buf += ptable_offset;
      size -= ptable_offset;
    }
    if (checksum(buf, size) != partition->csum) {
      fprintf(of, "Partition %d has bad checksum, was %lu should be %lu\n", 
              partition_no, checksum(buf, size), partition->csum);
      retval = -1;
    }
  }

  return retval;
}


static int
check_image_csum(const unsigned char *filebuf, int image_size)
{
  unsigned char hwid[HWID_SIZE+1];
  unsigned char csum[CSUM_SIZE+1];
  unsigned long checksum_value;

  memcpy(hwid, filebuf + image_size - HWID_SIZE - CSUM_SIZE, HWID_SIZE);
  hwid[HWID_SIZE] = '\0';

  memcpy(csum, filebuf + image_size - CSUM_SIZE, CSUM_SIZE);
  csum[CSUM_SIZE] = '\0';

  if (sscanf(csum, "%ld", &checksum_value) != 1) {
    fprintf(of, "Invalid image checksum '%s'\n", csum);
    return -1;
  }
  if (verbose)
    fprintf(of, "HWID %s, checksum %s (numeric %ld)\n", 
            hwid, csum, checksum_value);

  if (checksum(filebuf, image_size - CSUM_SIZE) != checksum_value) {
    fprintf(of, "Invalid checksum in image file\n");
    return -1;
  }
  return 0;
}


int
nandlib_check_image(const unsigned char *filebuf, 
                    int image_size, int start_partition)
{
  if (check_image_csum(filebuf, image_size) < 0)
    return -1;

  if (read_fimage_fwinfo(filebuf, &fwinfo) < 0) {
    fprintf(of, "Can't find image info area in image!\n");
    return -1;
  }

  if (read_partition_table(filebuf) < 0)
    return -1;

  if (start_partition >= nr_partitions) {
    fprintf(of, "Can't start at partition %d, only %d partitions\n",
            start_partition, nr_partitions);
    return -1;
  }

  if (check_partitions(filebuf) < 0)
    return -1;

  image_checked = 1;

  return 0;
}


int 
nandlib_check_flash(int fd)
{
  /* Fill in MTD device capability structure */  
  if (ioctl(fd, MEMGETINFO, &meminfo) != 0) {
    perror("MEMGETINFO");
    return -1;
  }

  if (verbose) {
    fprintf(of, "Flash total size 0x%x, erasesize %u, page size %u\n",
            meminfo.size, meminfo.erasesize, meminfo.writesize);
  }

  if (meminfo.type != MTD_NANDFLASH) {
    fprintf(of, "Flash type is not NAND. Aborting.\n");
    return -1;
  }

  flash_checked = 1;

  return 0;
}


int
nandlib_erase_flash(int fd, int start_partition, progressor progress)
{
  loff_t start_addr = partitions[start_partition].start_addr;
  loff_t flash_addr = start_addr;
  erase_info_t mtd_erase;

  if (!flash_checked) {
    fprintf(of, "Flash device not checked with nandlib_check_flash()!\n");
    return -1;
  }

  if (!image_checked) {
    fprintf(of, "Flash image not checked with nandlib_check_image()!\n");
    return -1;
  }

  while (flash_addr < meminfo.size) {
    int blockbad;

    blockbad = ioctl(fd, MEMGETBADBLOCK, &flash_addr);
    if (blockbad < 0) {
      perror("ioctl mtd MEMGETBADBLOCK");
      return -1;
    }

    if (blockbad && flash_addr < meminfo.erasesize) { /* block 0 bad */
      fprintf(of, "Block 0 bad. Aborting (system won't be able to boot).\n");
      return -1;
    }

    if (!blockbad) {
      mtd_erase.start = flash_addr;
      mtd_erase.length = meminfo.erasesize;
      if (ioctl(fd, MEMERASE, &mtd_erase) < 0) {
        if (errno == EIO) {
          /* Block became bad during write. Mark it, unless block 0. */
          if (flash_addr < meminfo.erasesize) { /* block 0 */
            fprintf(of, "Block 0 failed during erase, aborting.\n");
            return -1;
          }
          if (ioctl(fd, MEMSETBADBLOCK, &flash_addr) < 0) {
            perror("ioctl mtd MEMSETBADBLOCK");
            return -1;
          }
          if (verbose)
            fprintf(of, "Block at 0x%llx now bad, skipping\n", flash_addr);
        } else {
          perror("ioctl mtd MEMERASE");
          return -1;
        }
      }
    } else if (verbose) {
      fprintf(of, "Block at 0x%llx was bad (%d), skipping\n", 
              flash_addr, blockbad);
    }
    flash_addr += meminfo.erasesize;
    if (progress)
      indicate_progress(progress, 
                        flash_addr - start_addr, meminfo.size - start_addr);
  }
  return 0;
}


int
nandlib_program_flash(int fd, int start_partition, progressor progress)
{
  loff_t flash_addr = 0;
  int partition_no;
  int blockbad;
  int wrstatus;
  struct partition *partition;
  unsigned int len;
  const unsigned char *buf;
  const unsigned char *bufstart = NULL;
  uint32_t erasesize = meminfo.erasesize;

  if (!flash_checked) {
    fprintf(of, "Flash device not checked with nandlib_check_flash()!\n");
    return -1;
  }

  if (!image_checked) {
    fprintf(of, "Flash image not checked with nandlib_check_image()!\n");
    return -1;
  }

  /* Loop through all partitions, starting at the one specified */
  for (partition_no = start_partition; 
       partition_no < nr_partitions; 
       partition_no++) {
    partition = &partitions[partition_no];
    if (flash_addr > partition->start_addr) {
      fprintf(of, "Partition %d overlaps %d, aborting.\n", 
              partition_no, partition_no-1);
      return -1;
    }
    /* Fetch parameters from partition descriptor */
    flash_addr = partition->start_addr;
    len = partition->size;
    buf = partition->buf;
    if (!bufstart) /* 1st time */
      bufstart = buf;
    if (lseek(fd, flash_addr, SEEK_SET) < 0) {
      perror("seeking to partition start");
      return -1;
    }

    if (!quiet)
      fprintf(of, "Writing partition %d to 0x%llx...", 
              partition_no, flash_addr);

    /* Now write the whole partition */
    while (len > 0) {
      /* Check that we won't overrun the partition */
      if (flash_addr >= meminfo.size ||
          (partition_no < nr_partitions-1 && 
           flash_addr >= (partition+1)->start_addr) ||
          (partition_no == 0 && flash_addr >= fwinfo.bootblkpar_flash_addr)) {
        fprintf(of, "No room for partition %d, aborting.\n", partition_no);
        return -1;
      }

      /* Check for (already) bad block */
      blockbad = ioctl(fd, MEMGETBADBLOCK, &flash_addr);
      if (blockbad < 0) {
        perror("ioctl mtd MEMGETBADBLOCK");
        return -1;
      }

      if (blockbad && flash_addr < erasesize) { /* block 0 bad */
        fprintf(of, "Block 0 bad. Aborting (system won't be able to boot).\n");
        return -1;
      }

      /* Write if block was ok */
      if (!blockbad) {
        wrstatus = write(fd, buf, erasesize);
        if (wrstatus < 0) {
          if (errno == EIO) {
            /* Block became bad during write. Mark it, unless block 0. */
            if (flash_addr < erasesize) { /* block 0 */
              fprintf(of, "Block 0 failed during write, aborting.\n");
              return -1;
            }
            if (ioctl(fd, MEMSETBADBLOCK, &flash_addr) < 0) {
              perror("ioctl mtd MEMSETBADBLOCK");
              return -1;
            }
            blockbad = 1;
            if (verbose)
              fprintf(of, "Block at 0x%llx failed during write\n", flash_addr);
          } else {
            perror("writing block");
            return -1;
          }
        } else {
          /* write() successful, check length of write to be sure */
          if (wrstatus != erasesize) {
            fprintf(of, "Write failure at 0x%llx\n", flash_addr);
            return -1;
          }
          buf += erasesize;
          len -= erasesize;
        }
      }

      if (blockbad) {
        /* block was bad, lseek to new position to synchronize */
        if (lseek(fd, flash_addr + erasesize, SEEK_SET) < 0) {
          perror("skipping bad block during write");
          return -1;
        }
        if (verbose)
          fprintf(of, "Block at 0x%llx bad (%d), skipping\n", 
                 flash_addr, blockbad);
      }

      /* block was either written or bad; either way, advance flash addr */
      flash_addr += erasesize;
      if (progress)
        indicate_progress(progress, 
                          buf - bufstart, buf_max - bufstart);
    }
    if (!quiet)
      fprintf(of, "done.\n");
  }
  return 0;
}


int
nandlib_verify_flash(int fd, int start_partition, progressor progress)
{
  loff_t flash_addr = 0;
  int partition_no;
  int blockbad;
  int rdstatus;
  struct partition *partition;
  unsigned int len;
  const unsigned char *buf;
  const unsigned char *bufstart = NULL;
  unsigned char *blockbuf = NULL;
  uint32_t erasesize = meminfo.erasesize;
  int retval = -1; /* bad unless proven otherwise */

  if (!flash_checked) {
    fprintf(of, "Flash device not checked with nandlib_check_flash()!\n");
    return -1;
  }

  if (!image_checked) {
    fprintf(of, "Flash image not checked with nandlib_check_image()!\n");
    return -1;
  }

  if (!(blockbuf = malloc(erasesize))) {
    fprintf(of, "Can't allocate memory for verification block buffer\n");
    return -1;
  }

  /* Loop through all partitions, starting at the one specified */
  for (partition_no = start_partition; 
       partition_no < nr_partitions; 
       partition_no++) {
    partition = &partitions[partition_no];

    /* Fetch parameters from partition descriptor */
    flash_addr = partition->start_addr;
    len = partition->size;
    buf = partition->buf;
    if (!bufstart) /* 1st time */
      bufstart = buf;
    if (lseek(fd, flash_addr, SEEK_SET) < 0) {
      perror("seeking to partition start");
      goto exit;
    }

    if (!quiet)
      fprintf(of, "Verifying partition %d at 0x%llx...", partition_no, flash_addr);

    /* Now verify the whole partition */
    while (len > 0) {
      /* Check for bad block */
      blockbad = ioctl(fd, MEMGETBADBLOCK, &flash_addr);
      if (blockbad < 0) {
        perror("ioctl mtd MEMGETBADBLOCK");
        goto exit;
      }

      /* Verify if block was ok */
      if (!blockbad) {
        rdstatus = read(fd, blockbuf, erasesize);
        if (rdstatus < 0 || rdstatus != erasesize) {
          perror("reading block");
          goto exit;
        } else {
          /* read() successful, check content */
#ifdef FORCE_VERIFY_ERROR
          if (flash_addr > 0x130000)
            blockbuf[37] ^= 19;
#endif
          if (memcmp(buf, blockbuf, erasesize)) {
            fprintf(of, "Verify failure in block at 0x%llx\n", flash_addr);
            goto exit;
          }
          buf += erasesize;
          len -= erasesize;
        }
      }

      if (blockbad) {
        /* block was bad, lseek to new position to synchronize */
        if (lseek(fd, flash_addr + erasesize, SEEK_SET) < 0) {
          perror("skipping bad block during verify");
          goto exit;
        }
        if (verbose)
          fprintf(of, "Block at 0x%llx bad (%d), skipping\n", 
                 flash_addr, blockbad);
      }

      /* block was either ok or bad; either way, advance flash addr */
      flash_addr += erasesize;
      if (progress)
        indicate_progress(progress, 
                          buf - bufstart, buf_max - bufstart);
    }
    if (!quiet)
      fprintf(of, "done.\n");
  }
  retval = 0; /* all ok! */

exit:
  if (blockbuf)
    free(blockbuf);

  return retval;
}
