/* vi: set sw=2 ts=8: */
/*
 * nandlib - routines for NAND flash erase, write, partition checks
 *
 * (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef _NANDLIB_H_
#define _NANDLIB_H_

#define HWID_SIZE  (8)
#define CSUM_SIZE  (16)

#define MAX_ENTRIES (7) /* in partition table */

/* user function pointer type for progress indicator */
typedef void (*progressor)(unsigned int percent);

/* Set error/logging output file and verbosity level
 * verbose = 1 => very chatty, lots of info about partitions and bad blocks
 * quiet = 1 => very silent, only fatal errors are printed
 * verbose = quiet = 0 => roughly one line per partition printed
 */
void nandlib_set_verbosity(FILE *out_f, int new_verbose, int new_quiet);

/* Check image at filebuf, size image_size, starting with partition 
 * start_partition, save partition info for later.
 * Must be called prior to nandlib_program_flash.
 * Image must remain in memory until flash has been erased and programmed.
 * Return 0 if ok, < 0 if not (and print error message)
 */
int nandlib_check_image(const unsigned char *filebuf, 
                        int image_size, int start_partition);

/* Check flash for correct type, save flash geometry parameters for later.
 * Must be called prior to nandlib_erase_flash and nandlib_program_flash.
 * Return 0 if ok, < 0 if not (and print error message).
 */
int nandlib_check_flash(int fd);

/* Erase nand flash memory, starting with partition start_partition.
 * progress is user progress indicator function pointer, may be NULL.
 * Return 0 if all ok, < 0 if not (and print error message).
 */
int nandlib_erase_flash(int fd, int start_partition, progressor progress);

/* Program nand flash memory, starting with partition start_partition.
 * progress is user progress indicator function pointer, may be NULL.
 * Return 0 if all ok, < 0 if not (and print error message).
 */
int nandlib_program_flash(int fd, int start_partition, progressor progress);

/* Verify nand flash memory, starting with partition start_partition.
 * progress is user progress indicator function pointer, may be NULL.
 * Return 0 if all ok, < 0 if not (and print error message).
 */
int nandlib_verify_flash(int fd, int start_partition, progressor progress);

#endif /* _NANDLIB_H_ */
